/*
 * Author: Dušan Kolář
 */

#include "parse_goose.h"

#include <stddef.h>


#define GOOSE_MAGIC       0x88b8
#define VPN_MAGIC         0x8100
#define APDU_MAGIC        0x61
#define GOCB_REF_MAGIC    0x80
#define TTL_MAGIC         0x81
#define DATA_SET_MAGIC    0x82
#define GOOSE_ID_MAGIC    0x83
#define T_MAGIC           0x84
#define STATUS_NUM_MAGIC  0x85

#define MIN_LENGTH_APDU     25
#define MIN_LENGTH_GOOSE    35
#define MIN_LENGTH_ALL      47
        // GOOSE + 2xMAC(2x6B) + 2B_GOOSE_MAGIC
#define FCS_LENGTH           4


#define SWAP_TYPE(WIDTH) typedef union u_swap ## WIDTH {\
    uint ## WIDTH ## _t value;\
    uint8_t  bytes[WIDTH/8];\
} swap ## WIDTH ## _t;

SWAP_TYPE(64)
SWAP_TYPE(32)
SWAP_TYPE(16)


// .................................................................
// local function declaration

static goose_parse_error_t tlv_len(uint8_t **input, uint16_t *len, uint32_t *size);

// .................................................................
// local data

static goose_data_t data_res;

// .................................................................
// ------------------------ GLOBAL FUNCTIONS ------------------------


goose_data_t *parse_goose(uint8_t *input, uint32_t size, goose_parse_error_t *error) {
    // total length is too short to put in minimal information for GOOSE
    if (size < MIN_LENGTH_ALL) {
        *error = GPE_TOO_SHORT;
        return (NULL);
    }
    
    input += 12; // skip 2 MAC addresses
    
    swap16_t goose_magic;
    goose_magic.bytes[1] = *input++;
    goose_magic.bytes[0] = *input++;
    
    if (goose_magic.value != GOOSE_MAGIC) {
        if (goose_magic.value != VPN_MAGIC) {
            *error = GPE_GOOSE_MAGIC_FAIL;
            return (NULL);
        }
        size -=  4;
        input += 2;
        goose_magic.bytes[1] = *input++;
        goose_magic.bytes[0] = *input++;
        if (goose_magic.value != GOOSE_MAGIC) {
            *error = GPE_GOOSE_MAGIC_FAIL;
            return (NULL);
        }
    }
    
    size -= 14; // processed 14 bytes;

    // APP ID, GOOSE LEN PARSE
    
    swap16_t appid;
    appid.bytes[1] = *input++;
    appid.bytes[0] = *input++;
    data_res.appid = appid.value;

    swap16_t len_goose;
    len_goose.bytes[1] = *input++;
    len_goose.bytes[0] = *input++;
    
    if (len_goose.value != size) {
        // length mismatch
        *error = GPE_LENGTH_MISMATCH;
        return (NULL);
    }
    
    size -= 4;  // 2 bytes of APPID processed, 2 bytes of length processed

    input += 4; // skip 4 reserved bytes
    size  -= 4;
    
    // ADPU MAGIC & LENGTH PARSE
    
    if (*input++ != APDU_MAGIC) {
        *error = GPE_APDU_MAGIC_FAIL;
        return (NULL);
    }
    --size;
    
    goose_parse_error_t err;
    uint16_t tmp_len;
    
    if ((err = tlv_len(&input, &tmp_len, &size)) != GPE_OK) {
        *error = err;
        return (NULL);
    }

    if (tmp_len != size) {
        // length mismatch, remaining length and data size are different
        *error = GPE_LENGTH_MISMATCH;
        return (NULL);
    }
    
    // GOCB REF PARSE
    
    if (*input++ != GOCB_REF_MAGIC) {
        *error = GPE_GOOSE_MAGIC_FAIL;
        return (NULL);
    }
    --size;
    
    if ((err = tlv_len(&input, &tmp_len, &size)) != GPE_OK) {
        *error = err;
        return (NULL);
    }

    data_res.gocb_ref_len  = tmp_len;
    data_res.gocb_ref_data = input;
    input += tmp_len;
    
    if (tmp_len > size) {
        // string length higher than packet size
        *error = GPE_LENGTH_MISMATCH;
        return (NULL);
    }
    size -= tmp_len;

    // TTL SKIP
    
    if (size < 2) {
        // packet size too low
        *error = GPE_LENGTH_MISMATCH;
        return (NULL);
    }
    if (*input++ != TTL_MAGIC) {
        *error = GPE_TTL_MAGIC_FAIL;
        return (NULL);
    }
    --size;

    if ((err = tlv_len(&input, &tmp_len, &size)) != GPE_OK) {
        *error = err;
        return (NULL);
    }

    input += tmp_len;

    if (tmp_len > size) {
        // string length higher than packet size
        *error = GPE_LENGTH_MISMATCH;
        return (NULL);
    }
    size -= tmp_len;

    // DATA SET PARSE
    
    if (size < 2) {
        // packet size too low
        *error = GPE_LENGTH_MISMATCH;
        return (NULL);
    }
    if (*input++ != DATA_SET_MAGIC) {
        *error = GPE_DATA_SET_MAGIC_FAIL;
        return (NULL);
    }
    --size;

    if ((err = tlv_len(&input, &tmp_len, &size)) != GPE_OK) {
        *error = err;
        return (NULL);
    }

    data_res.data_set_len  = tmp_len;
    data_res.data_set_data = input;
    input += tmp_len;

    if (tmp_len > size) {
        // string length higher than packet size
        *error = GPE_LENGTH_MISMATCH;
        return (NULL);
    }
    size -= tmp_len;

    // GOOSE ID PARSE
    
    if (size < 2) {
        // packet size too low
        *error = GPE_LENGTH_MISMATCH;
        return (NULL);
    }
    if (*input++ != GOOSE_ID_MAGIC) {
        *error = GPE_GOOSE_ID_MAGIC_FAIL;
        return (NULL);
    }
    --size;

    if ((err = tlv_len(&input, &tmp_len, &size)) != GPE_OK) {
        *error = err;
        return (NULL);
    }

    data_res.goose_id_len  = tmp_len;
    data_res.goose_id_data = input;
    input += tmp_len;

    if (tmp_len > size) {
        // string length higher than packet size
        *error = GPE_LENGTH_MISMATCH;
        return (NULL);
    }
    size -= tmp_len;

    // T SKIP
    
    if (size < 2) {
        // packet size too low
        *error = GPE_LENGTH_MISMATCH;
        return (NULL);
    }
    if (*input++ != T_MAGIC) {
        *error = GPE_T_MAGIC_FAIL;
        return (NULL);
    }
    --size;

    if ((err = tlv_len(&input, &tmp_len, &size)) != GPE_OK) {
        *error = err;
        return (NULL);
    }

    input += tmp_len;

    if (tmp_len > size) {
        // string length higher than packet size
        *error = GPE_LENGTH_MISMATCH;
        return (NULL);
    }
    size -= tmp_len;

    // STATUS NUM PARSE
    
    if (size < 2) {
        // packet size too low
        *error = GPE_LENGTH_MISMATCH;
        return (NULL);
    }
    if (*input++ != STATUS_NUM_MAGIC) {
        *error = GPE_STATUS_NUM_MAGIC_FAIL;
        return (NULL);
    }
    --size;

    if ((err = tlv_len(&input, &tmp_len, &size)) != GPE_OK) {
        *error = err;
        return (NULL);
    }

    data_res.status_num_len   = tmp_len;
    data_res.status_num_data  = input;
    data_res.status_num_value = 0;

    if (tmp_len > size) {
        // string length higher than packet size
        *error = GPE_LENGTH_MISMATCH;
        return (NULL);
    }
    
    if (tmp_len<=4) { // should be always > 1
        swap32_t tmp32;
        tmp32.value = 0;
        uint16_t s = tmp_len;
        uint16_t i = 0;
        while (s>0) {
            tmp32.bytes[--s] = input[i++];
        }
        data_res.status_num_value = tmp32.value;
    }
    input += tmp_len;
    size -= tmp_len;
    
    *error = GPE_OK;
    return (&data_res);
}


// .................................................................
// ------------------------ LOCAL FUNCTIONS ------------------------


static goose_parse_error_t tlv_len(uint8_t **input, uint16_t *len, uint32_t *size) {
    uint8_t *in = *input;
    uint8_t enc = *in++;

    if (*size < 1) {
        *len   = 0;
        return (GPE_OUT_OF_DATA);
    }
    
    if (enc<128) { // short len, just this one byte carries length information
        *input = in;
        *len   = enc;
        --(*size);  // just one encoding byte read
        return (GPE_OK);
    }
    enc = enc & 0x7f;
    if (enc == 0) { // indefinite length, failure
        *input = in;
        *len   = 0;
        return (GPE_TLV_INDEFINITE);
    }
    if (enc > 2) { // number of length bytes > 2, probably error
        *input = in;
        *len = 0;
        return (GPE_TLV_LARGE);
    }
    
    if (*size < 2) {
        *input = in;
        *len   = 0;
        return (GPE_OUT_OF_DATA);
    }
    
    if (enc == 1) {
        *len = (uint16_t)(*in++);
        *input = in;
        (*size) -= 2;  // encoding and 1 length byte
        return (GPE_OK);
    }
    
    if (*size < 3) {
        *input = in;
        *len   = 0;
        return (GPE_OUT_OF_DATA);
    }
    
    // remains length 2, should change BE -> LE
    swap16_t sw_len;
    sw_len.bytes[1] = *in++;
    sw_len.bytes[0] = *in++;
    *len = sw_len.value;
    *input = in;
    (*size) -= 3;  // encoding and 2 length bytes
    return (GPE_OK);
}


// EOF parse_goose
// .................................................................
// .................................................................
