/*
 * Author: Dušan Kolář
 */

#include <stdio.h>
#include <stdlib.h>
#include <pcap.h>
#include <errno.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>
#include <arpa/inet.h>
#include <netinet/if_ether.h> 
#include <err.h>
#include <inttypes.h>

#include "parse_goose.h"

#ifdef __linux__                 // for Linux
#include <netinet/ether.h> 
#include <time.h>
#include <pcap/pcap.h>
#endif

#ifndef PCAP_ERRBUF_SIZE
#define PCAP_ERRBUF_SIZE (256)
#endif

#define SIZE_ETHERNET (14)       // offset of Ethernet header to L3 protocol


static void print_data_str(uint16_t len, uint8_t *ptr) {
    while (len>0) {
        putchar(*ptr++);
        --len;
    }
}

static void print_data_hex(uint16_t len, uint8_t *ptr) {
    while (len>0) {
        printf("%02X",*ptr++);
        --len;
    }
}

static void process_goose(uint8_t *inp, uint32_t size) {
    goose_parse_error_t error;
    goose_data_t *res = parse_goose(inp, size, &error);
    if (res != NULL) {
        printf("APPID: %u\n", res->appid);
        
        printf("GOOSE control block: ");
        print_data_str( res->gocb_ref_len, res->gocb_ref_data );
        puts("");
        
        printf("Data set: ");
        print_data_str( res->data_set_len, res->data_set_data );
        puts("");
        
        printf("GOOSE ID: ");
        print_data_str( res->goose_id_len, res->goose_id_data );
        puts("");
        
        printf("Status number: ");
        print_data_hex( res->status_num_len, res->status_num_data );
        puts("");
        printf("Status number value: %u\n",res->status_num_value);
            
    } else {
        // error
        switch (error) {
            case GPE_OK:
                puts("Everything was OK, strange...");
                break;
            case GPE_TOO_SHORT:                // passed buffer is shorter then smallest packet
                puts("passed buffer is shorter then smallest packet");
                break;
            case GPE_LENGTH_MISMATCH:          // lengths inside don't match
                puts("lengths inside don't match");
                break;
            case GPE_OUT_OF_DATA:              // running out of the passed data length
                puts("running out of the passed data length");
                break;
            case GPE_TLV_LARGE:                // size of TLV goes more than 0xffff
                puts("size of TLV goes more than 0xffff");
                break;
            case GPE_TLV_INDEFINITE:           // TLV length set as indefinite
                puts("TLV length set as indefinite");
                break;
    
            case GPE_GOOSE_MAGIC_FAIL:         // GOOSE magic expected, not found
                puts("GOOSE magic expected, not found");
                break;
            case GPE_APDU_MAGIC_FAIL:          // APDU magic expected, not found
                puts("APDU magic expected, not found");
                break;
            case GPE_GOCB_REF_MAGIC_FAIL:      // gocbRef magic expected, not found
                puts("gocbRef magic expected, not found");
                break;
            case GPE_TTL_MAGIC_FAIL:           // Time to live magic expected, not found
                puts("Time to live magic expected, not found");
                break;
            case GPE_DATA_SET_MAGIC_FAIL:      // Data set magic expected, not found
                puts("Data set magic expected, not found");
                break;
            case GPE_GOOSE_ID_MAGIC_FAIL:      // GOOSE ID magic expected, not found
                puts("GOOSE ID magic expected, not found");
                break;
            case GPE_T_MAGIC_FAIL:             // T magic expected, not found
                puts("T magic expected, not found");
                break;
            case GPE_STATUS_NUM_MAGIC_FAIL:    // Status number magic expected, not found
                puts("Status number magic expected, not found");
                break;
            default:
                puts("Strange goose error...");
                break;
        }  // of switch (error)
    }
}

goose_data_t *parse_goose(uint8_t *input, uint32_t size, goose_parse_error_t *error);

int main(int argc, char *argv[]){
    int n;
    char errbuf[PCAP_ERRBUF_SIZE];  // constant defined in pcap.h
    const u_char *packet;
    struct ip *my_ip;
    const struct tcphdr *my_tcp;    // pointer to the beginning of TCP header
    const struct udphdr *my_udp;    // pointer to the beginning of UDP header
    struct pcap_pkthdr header;  
    struct ether_header *eptr;
    pcap_t *handle;                 // file/device handler
    u_int size_ip;
  
    if (argc != 2) {                 // one parameter expected
        errx(1,"Usage: %s <filename>",argv[0]);
    }
  
    // open the input file
    if ((handle = pcap_open_offline(argv[1],errbuf)) == NULL) {
        err(1,"Can't open file %s for reading",argv[1]);
    }
  
    printf("Opening file %s for reading ...\n\n", argv[1]);
    n = 0;
  
    // read packets from the file
    while ((packet = pcap_next(handle,&header)) != NULL) {
        n++;
        // print the packet header data
        printf("Packet no. %d:\n",n);
        printf("\tLength %d, received at %s",header.len,ctime((const time_t*)&header.ts.tv_sec));  

        // read the Ethernet header
        eptr = (struct ether_header *) packet;
        printf("\tSource MAC: %s\n",ether_ntoa((const struct ether_addr *)&eptr->ether_shost)) ;
        printf("\tDestination MAC: %s\n",ether_ntoa((const struct ether_addr *)&eptr->ether_dhost)) ;
    
        switch (ntohs(eptr->ether_type)) {               // see /usr/include/net/ethernet.h for types
            case ETHERTYPE_IP: // IPv4 packet
                printf("\tEthernet type is  0x%x, i.e. IP packet \n", ntohs(eptr->ether_type));
                my_ip = (struct ip*) (packet+SIZE_ETHERNET);        // skip Ethernet header
                size_ip = my_ip->ip_hl*4;                           // length of IP header

                printf("\tIP: id 0x%x, hlen %d bytes, version %d, total length %d bytes, TTL %d\n",
                            ntohs(my_ip->ip_id), size_ip, my_ip->ip_v, ntohs(my_ip->ip_len), my_ip->ip_ttl);
                printf("\tIP src = %s, ",inet_ntoa(my_ip->ip_src));
                printf("IP dst = %s",inet_ntoa(my_ip->ip_dst));

                switch (my_ip->ip_p) {
                    case 2: // IGMP protocol
                        printf(", protocol IGMP (%d)\n",my_ip->ip_p);
                        break;
                        
                    case 6: // TCP protocol
                        printf(", protocol TCP (%d)\n",my_ip->ip_p);
                        my_tcp = (struct tcphdr *) (packet+SIZE_ETHERNET+size_ip); // pointer to the TCP header
                        printf("\tSrc port = %d, dst port = %d, seq = %u",
                                    ntohs(my_tcp->th_sport), ntohs(my_tcp->th_dport), ntohl(my_tcp->th_seq));
                        if (my_tcp->th_flags & TH_SYN) {
                            printf(", SYN");
                        }
                        if (my_tcp->th_flags & TH_FIN) {
                            printf(", FIN");
                        }
                        if (my_tcp->th_flags & TH_RST) {
                            printf(", RST");
                        }
                        if (my_tcp->th_flags & TH_PUSH) {
                            printf(", PUSH");
                        }
                        if (my_tcp->th_flags & TH_ACK) {
                            printf(", ACK");
                        }
                        printf("\n");
                        break;
                        
                    case 17: // UDP protocol
                        printf(", protocol UDP (%d)\n",my_ip->ip_p);
                        my_udp = (struct udphdr *) (packet+SIZE_ETHERNET+size_ip); // pointer to the UDP header
                        printf("\tSrc port = %d, dst port = %d, length %d\n",
                                    ntohs(my_udp->uh_sport), ntohs(my_udp->uh_dport), ntohs(my_udp->uh_ulen));
                        break;
                        
                    default: 
                        printf(", protocol %d\n",my_ip->ip_p);
                } // of switch (my_ip->ip_p)
                break;
                
            case ETHERTYPE_IPV6:  // IPv6
                printf("\tEthernet type is 0x%x, i.e., IPv6 packet\n",ntohs(eptr->ether_type));
                break;
                
            case ETHERTYPE_ARP:  // ARP
                printf("\tEthernet type is 0x%x, i.e., ARP packet\n",ntohs(eptr->ether_type));
                break;
                
            default:
                process_goose((uint8_t *)packet,(uint32_t)header.len);
                break;
        } // of switch (ntohs(eptr->ether_type))
    
        puts("");
    }  // while
    
    puts("End of file reached ...");
  
    // close the capture device and deallocate resources
    pcap_close(handle);
    
    return 0;
}

// EOF
